from ansi2html import Ansi2HTMLConverter
from django.shortcuts import get_object_or_404
from rest_framework import generics, viewsets
from rest_framework.decorators import api_view
from rest_framework.response import Response

from api.models import SaltReturns, SaltEvents, Jids
from api.serializers import SaltReturnsSerializer, EventsSerializer
from api.utils.output import highstate_output, nested_output


class MultipleFieldLookupMixin(object):
    """
    Apply this mixin to any view or viewset to get multiple field filtering
    based on a `lookup_fields` attribute, instead of the default single field filtering.
    """

    def get_object(self):
        queryset = self.get_queryset()  # Get the base queryset
        queryset = self.filter_queryset(queryset)  # Apply any filter backends
        filter = {}
        for field in self.lookup_fields:
            if self.kwargs[field]:  # Ignore empty fields.
                filter[field] = self.kwargs[field]
        obj = get_object_or_404(queryset, **filter)  # Lookup the object
        self.check_object_permissions(self.request, obj)
        return obj


class SaltReturnsList(generics.ListAPIView):
    serializer_class = SaltReturnsSerializer

    def get_queryset(self):
        queryset = SaltReturns.objects.all()
        qry = {}
        start = self.request.query_params.get("start", None)
        end = self.request.query_params.get("end", None)
        limit = int(self.request.query_params.get("limit", 200))
        target = self.request.query_params.getlist("target[]")
        users = self.request.query_params.getlist("users[]", None)
        if target:
            if len(target) > 1:
                qry["id__in"] = target
            else:
                qry["id"] = target[0]
        if start and end:
            qry["alter_time__date__range"] = [start, end]

        queryset = queryset.filter(**qry).order_by("-alter_time")[:limit]
        if users:
            queryset = [i for i in queryset if i.user() in users]
        return queryset


@api_view(["GET"])
def jobs_filters(request):
    # Filter options.
    user_list = list({i.user() for i in Jids.objects.all()})
    minion_list = SaltReturns.objects.values_list("id", flat=True).distinct()
    return Response({"users": user_list, "minions": minion_list})


class SaltReturnsRetrieve(MultipleFieldLookupMixin, generics.RetrieveAPIView):
    queryset = SaltReturns.objects.all().order_by("-alter_time")
    serializer_class = SaltReturnsSerializer
    lookup_fields = ["jid", "id"]


@api_view(["GET"])
def job_rendered(request, jid, minion_id):
    # Retrieve job from database.
    job = SaltReturns.objects.get(jid=jid, id=minion_id)

    # Use different output.
    if job.fun in ["state.apply", "state.highstate"]:
        formatted = highstate_output.output({minion_id: job.loaded_ret()["return"]})
    else:
        formatted = nested_output.output({minion_id: job.loaded_ret()["return"]})

    # Convert it to html.
    conv = Ansi2HTMLConverter(inline=False, scheme="xterm")
    html_detail = conv.convert(formatted, ensure_trailing_newline=True)
    return Response(html_detail)


class EventsViewSet(viewsets.ReadOnlyModelViewSet):
    """
    A simple ViewSet for viewing accounts.
    """

    queryset = SaltEvents.objects.all().order_by("-alter_time")[:200]
    serializer_class = EventsSerializer



from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand, CommandError


UserModel = get_user_model()


class Command(BaseCommand):
    help = "Create Token for a given user"

    def create_user_token(self, username, reset_token):
        user = UserModel.objects.get(username=username)

        # Just saving the user should be enough.
        if reset_token:
            user.user_settings.generate_token()

        token = UserModel.objects.get(username=username).user_settings.token
        return token

    def add_arguments(self, parser):
        parser.add_argument("username", type=str)

        parser.add_argument(
            "-r",
            "--reset",
            action="store_true",
            dest="reset_token",
            default=False,
            help="Reset existing User token and create a new one",
        )

    def handle(self, *args, **options):
        username = options["username"]
        reset_token = options["reset_token"]

        try:
            token = self.create_user_token(username, reset_token)
        except UserModel.DoesNotExist:
            raise CommandError(
                "Cannot create the Token: user {0} does not exist".format(username)
            )
        if options["reset_token"]:
            self.stdout.write(
                "Generated token {0} for user {1}".format(token, username)
            )
        else:
            self.stdout.write("user {1}'s token: {0}".format(token, username))



import os

import MySQLdb
from django.conf import settings
from django.db import connections
from django.contrib.auth.models import User
from django.core.management import call_command
from django.test.client import Client

from .fixtures import *


@pytest.fixture()
def run_sql():
    def _run_sql(sql):
        conn = MySQLdb.connect(
            host=os.environ["DB_HOST"],
            user=os.environ["DB_USER"],
            passwd=os.environ["DB_PASS"],
            db=os.environ["DB_NAME"],
            port=int(os.environ["DB_PORT"]),
            ssl={},
        )
        cursor = conn.cursor()
        cursor.execute(sql)
        conn.commit()

    return _run_sql


# noinspection SqlNoDataSourceInspection
@pytest.yield_fixture(scope="session", autouse=True)
def django_db_setup(django_db_blocker):
    settings.DATABASES["default"] = {
        "ENGINE": "django.db.backends.mysql",
        "HOST": os.environ["DB_HOST"],
        "NAME": os.environ["DB_NAME"],
        "PORT": int(os.environ["DB_PORT"]),
    }
    settings.USE_TZ = False


# # noinspection SqlNoDataSourceInspection
# @pytest.yield_fixture(scope="session", autouse=True)
# def django_db_setup(django_db_blocker):
#     settings.DATABASES["default"] = {
#         "ENGINE": "django.db.backends.mysql",
#         "HOST": os.environ["DB_HOST"],
#         "NAME": os.environ["DB_NAME"],
#         "PORT": int(os.environ["DB_PORT"]),
#     }
#     settings.USE_TZ = False
#
#     yield
#
#     for connection in connections.all():
#         connection.close()
#
#     conn = MySQLdb.connect(
#         host=os.environ["DB_HOST"],
#         user=os.environ["DB_USER"],
#         passwd=os.environ["DB_PASS"],
#         db=os.environ["DB_NAME"],
#         port=int(os.environ["DB_PORT"]),
#         ssl={},
#     )
#     cursor = conn.cursor()
#     cursor.execute("DROP DATABASE IF EXISTS  `salt`")
#     cursor.execute("CREATE DATABASE `salt`")
#     cursor.execute("USE `salt`")
#     cursor.execute(
#         """CREATE TABLE `jids` (
#                         `jid` varchar(255) NOT NULL,
#                         `load` mediumtext NOT NULL,
#                         UNIQUE KEY `jid` (`jid`)
#                       ) ENGINE=InnoDB DEFAULT CHARSET=utf8"""
#     )
#
#     cursor.execute(
#         """CREATE TABLE `salt_returns` (
#                         `fun` varchar(50) NOT NULL,
#                         `jid` varchar(255) NOT NULL,
#                         `return` mediumtext NOT NULL,
#                         `id` varchar(255) NOT NULL,
#                         `success` varchar(10) NOT NULL,
#                         `full_ret` mediumtext NOT NULL,
#                         `alter_time` TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
#                         KEY `id` (`id`),
#                         KEY `jid` (`jid`),
#                         KEY `fun` (`fun`)
#                       ) ENGINE=InnoDB DEFAULT CHARSET=utf8"""
#     )
#
#     cursor.execute(
#         """CREATE TABLE `salt_events` (
#                         `id` BIGINT NOT NULL AUTO_INCREMENT,
#                         `tag` varchar(255) NOT NULL,
#                         `data` mediumtext NOT NULL,
#                         `alter_time` TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
#                         `master_id` varchar(255) NOT NULL,
#                         PRIMARY KEY (`id`),
#                         KEY `tag` (`tag`)
#                       ) ENGINE=InnoDB DEFAULT CHARSET=utf8"""
#     )
#     conn.commit()
#
#     with django_db_blocker.unblock():
#         call_command("migrate", verbosity=0, interactive=False)
#         User.objects.filter(username="admin").count() or User.objects.create_superuser(
#             "admin", "admin@example.com", "password"
#         )
#         client = Client()
#         username = "admin"
#         password = "password"
#         client.post("/api/token/", {"username": username, "password": password})
#         client.get("/", follow=True)
#         client.post("/run", {"action": "reject", "target": "*"})



import os
from django.core.management.base import BaseCommand
from django.db import connections
from django.db.utils import OperationalError


class Command(BaseCommand):
    help = "Check access to database and that env var are set"

    def handle(self, *args, **options):

        unset = []
        for env in [
            "MASTER_MINION_ID",
            "DB_BACKEND",
            "DB_NAME",
            "DB_USER",
            "DB_PASS",
            "DB_HOST",
            "DB_PORT",
            "SECRET_KEY",
            "ALLOWED_HOSTS",
            "SALT_URL",
            "SALT_AUTH",
        ]:
            try:
                os.environ[env]
            except KeyError:
                unset.append(env)

        db_conn = connections["default"]
        try:
            db_conn.cursor()
        except OperationalError as e:
            error = e
        else:
            error = None

        self.stdout.write("db:\t{}\nenv:\t{}".format(error or "ok", unset or "ok"))



from django.urls import path, include
from rest_framework_simplejwt.views import TokenRefreshView

from api.views.salt import (
    SaltReturnsList,
    SaltReturnsRetrieve,
    EventsViewSet,
    jobs_filters,
    job_rendered,
)

from api.views.alcali import (
    index_view,
    KeysViewSet,
    MinionsViewSet,
    jobs_graph,
    stats,
    event_stream,
    parse_modules,
    ConformityViewSet,
    UsersViewSet,
    MinionsCustomFieldsViewSet,
    FunctionsViewSet,
    run,
    UserSettingsViewSet,
    ScheduleViewSet,
    MyTokenObtainPairView,
    search,
)
from rest_framework import routers

router = routers.DefaultRouter()
router.register(r"keys", KeysViewSet)
router.register(r"minions", MinionsViewSet)
router.register(r"events", EventsViewSet)
router.register(r"conformity", ConformityViewSet)
router.register(r"users", UsersViewSet)
router.register(r"userssettings", UserSettingsViewSet)
router.register(r"minionsfields", MinionsCustomFieldsViewSet)
router.register(r"functions", FunctionsViewSet)
router.register(r"schedules", ScheduleViewSet)

urlpatterns = [
    path("", index_view, name="index"),
    path("api/token/", MyTokenObtainPairView.as_view(), name="token_obtain_pair"),
    path("api/token/refresh/", TokenRefreshView.as_view(), name="token_refresh"),
    path("api/", include(router.urls)),
    path("api/search/", search, name="search"),
    path("api/stats/", stats, name="stats"),
    path("api/settings/initdb", parse_modules, name="parse_modules"),
    path("api/event_stream/", event_stream, name="event_stream"),
    path("api/jobs/", SaltReturnsList.as_view(), name="jobs-list"),
    path("api/jobs/filters/", jobs_filters, name="jobs-filters"),
    path("api/run/", run, name="run"),
    path(
        "api/jobs/<str:jid>/<str:id>/",
        SaltReturnsRetrieve.as_view(),
        name="jobs-detail",
    ),
    path(
        "api/jobs/<str:jid>/<str:minion_id>/rendered_state/",
        job_rendered,
        name="jobs-detail-rendered",
    ),
    path("api/jobs/graph", jobs_graph, name="jobs_graph"),
]



from datetime import timedelta
import pytest

from django.utils import timezone

from api.utils import RawCommand, graph_data, render_conformity


def test_simple_command():
    command = RawCommand("salt '*' cmd.run 'echo foobar'")
    parsed = command.parse()
    assert parsed[0]["tgt_type"] == "glob"
    assert parsed[0]["tgt"] == "*"
    assert parsed[0]["fun"] == "cmd.run"
    assert parsed[0]["client"] == "local"
    assert parsed[0]["arg"] == ["echo foobar"]


def test_simple_command_inline():
    command = RawCommand("'*' cmd.run 'echo foobar'", inline=True)
    parsed = command.parse()
    assert parsed[0]["tgt_type"] == "glob"
    assert parsed[0]["tgt"] == "*"
    assert parsed[0]["fun"] == "cmd.run"
    assert parsed[0]["client"] == "local"
    assert parsed[0]["arg"] == ["echo foobar"]


def test_complex_command():
    command = RawCommand("salt -G 'os:Fedora' cmd.run 'echo foobar' test=True")
    parsed = command.parse()
    assert parsed[0]["tgt_type"] == "grain"
    assert parsed[0]["tgt"] == "os:Fedora"
    assert parsed[0]["fun"] == "cmd.run"
    assert parsed[0]["client"] == "local"
    assert parsed[0]["arg"] == ["echo foobar", "test=True"]


def test_client():
    command = RawCommand("salt --client=runner pillar.show_top")
    parsed = command.parse()
    assert parsed[0]["fun"] == "pillar.show_top"
    assert parsed[0]["client"] == "runner"

    command = RawCommand("salt --client=runner pillar.show_top master")
    parsed = command.parse()
    assert parsed[0] == {
        "fun": "pillar.show_top",
        "client": "runner",
        "arg": ["master"],
    }

    command = RawCommand("salt --client=wheel key.list_all")
    parsed = command.parse()
    assert parsed[0] == {"fun": "key.list_all", "client": "wheel"}

    command = RawCommand("salt --client=wheel key.finger master")
    parsed = command.parse()
    assert parsed[0] == {"fun": "key.finger", "client": "wheel", "arg": ["master"]}

    command = RawCommand("salt --client=foobar key.finger master")
    parsed = command.parse()
    assert parsed == "Client not implemented: foobar"


def test_failed_cmd():
    command = RawCommand("salt '*'")
    parsed = command.parse()
    assert parsed == "Command or target not specified"


def test_batch():
    command = RawCommand("salt --client=local_batch '*' state.apply test=True -b 1")
    parsed = command.parse()
    assert parsed[0]["batch"] == "1"


def test_timeout():
    command = RawCommand("salt --client=local '*' state.apply test=True -t 9")
    parsed = command.parse()
    assert parsed[0]["timeout"] == 9


def test_client_kwarg():
    command = RawCommand("salt --client=runner pillar.show_top kwarg='{foo: bar}'")
    parsed = command.parse()
    assert parsed[0]["fun"] == "pillar.show_top"
    assert parsed[0]["client"] == "runner"
    assert "kwarg" in parsed[0]

    command = RawCommand("salt --client=wheel key.list_all kwarg='{foo: bar}'")
    parsed = command.parse()
    assert parsed[0] == {
        "fun": "key.list_all",
        "client": "wheel",
        "kwarg": "{foo: bar}",
    }


@pytest.mark.django_db
def test_graph_data():
    today = str(timezone.now().date() - timedelta(days=2))
    days, count, error_count = graph_data(period=3, fun="all")
    assert today in days


@pytest.mark.django_db
def test_render_conformity(minion_master, foo_conformity, dummy_state):
    conformity_names, ret, details = render_conformity()
    assert conformity_names == ["FOO"]
    assert ret == [{"noice": 1}]
    assert details == {"master": [{"foo": "noice"}]}



from shlex import split
import json
from json import JSONDecodeError


class RawCommand:
    def __init__(self, command, client="local", posix=True, inline=False):
        # TODO: check shlex.quote, raw string, etc..
        if inline:
            self.command = split(command, posix=posix)
        else:
            self.command = split(command, posix=posix)[1:]
        self.options = {"expr_form": "glob"}
        self.client = client

    def parse(self):
        args = self.command

        if args[0].startswith("--client"):
            self.client = args[0].split("=")[1]
            args.pop(0)

        low = {"client": self.client}

        if self.client.startswith("local"):
            if len(args) < 2:
                return "Command or target not specified"

            # Batch option
            low["batch"] = None
            if self.client == "local_batch":
                batch_index = None
                for index, arg in enumerate(args):
                    if arg in ["-b", "--batch", "--batch-size"]:
                        low["batch"] = args[index + 1]
                        batch_index = index
                if batch_index:
                    args.pop(batch_index)
                    args.pop(batch_index)
            # Timeout option
            timeout_index = None
            for index, arg in enumerate(args):
                if arg in ["-t", "--timeout"]:
                    low["timeout"] = int(args[index + 1])
                    timeout_index = index
            if timeout_index:
                args.pop(timeout_index)
                args.pop(timeout_index)

            # take care of targeting.
            target_dict = {
                "pcre": ["-E", "--pcre"],
                "list": ["-L", "--list"],
                "grain": ["-G", "--grain"],
                "grain_pcre": ["--grain-pcre"],
                "pillar": ["-I", "--pillar"],
                "pillar_pcre": ["--pillar-pcre"],
                "range": ["-R", "--range"],
                "compound": ["-C", "--compound"],
                "nodegroup": ["-N", "--nodegroup"],
            }
            for key, value in target_dict.items():
                if args[0] in value:
                    self.options["expr_form"] = key
                    args.pop(0)

            low["tgt_type"] = self.options["expr_form"]
            low["tgt"] = args.pop(0)
            low["fun"] = args.pop(0)
            low["arg"] = args

        elif self.client.startswith("runner") or self.client.startswith("wheel"):
            low["fun"] = args.pop(0)
            for arg in args:
                if "=" in arg:
                    key, value = arg.split("=", 1)
                    try:
                        low[key] = json.loads(value)
                    except JSONDecodeError:
                        low[key] = value
                else:
                    low.setdefault("arg", []).append(arg)

        else:
            # This should never happen
            return "Client not implemented: {0}".format(self.client)

        return [low]



# Generated by Django 2.2.4 on 2019-08-17 19:37

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("auth", "0011_update_proxy_permissions"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Jids",
            fields=[
                (
                    "jid",
                    models.CharField(
                        db_index=True, max_length=255, primary_key=True, serialize=False
                    ),
                ),
                ("load", models.TextField()),
            ],
            options={"db_table": "jids", "managed": False},
        ),
        migrations.CreateModel(
            name="SaltEvents",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                ("tag", models.CharField(db_index=True, max_length=255)),
                ("data", models.TextField()),
                ("alter_time", models.DateTimeField()),
                ("master_id", models.CharField(max_length=255)),
            ],
            options={"db_table": "salt_events", "managed": False},
        ),
        migrations.CreateModel(
            name="SaltReturns",
            fields=[
                ("fun", models.CharField(db_index=True, max_length=50)),
                ("jid", models.CharField(db_index=True, max_length=255)),
                ("return_field", models.TextField(db_column="return")),
                (
                    "id",
                    models.CharField(max_length=255, primary_key=True, serialize=False),
                ),
                ("success", models.CharField(max_length=10)),
                ("full_ret", models.TextField()),
                ("alter_time", models.DateTimeField()),
            ],
            options={"db_table": "salt_returns", "managed": False},
        ),
        migrations.CreateModel(
            name="Conformity",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255)),
                ("function", models.CharField(max_length=255)),
            ],
            options={"db_table": "conformity"},
        ),
        migrations.CreateModel(
            name="Functions",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255)),
                ("type", models.CharField(max_length=255)),
                ("description", models.TextField()),
            ],
            options={"db_table": "salt_functions"},
        ),
        migrations.CreateModel(
            name="Keys",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("minion_id", models.CharField(max_length=255)),
                ("pub", models.TextField(blank=True)),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("accepted", "accepted"),
                            ("rejected", "rejected"),
                            ("denied", "denied"),
                            ("unaccepted", "unaccepted"),
                        ],
                        max_length=64,
                    ),
                ),
            ],
            options={"db_table": "salt_keys"},
        ),
        migrations.CreateModel(
            name="Minions",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("minion_id", models.CharField(max_length=128)),
                ("grain", models.TextField()),
                ("pillar", models.TextField()),
            ],
            options={"db_table": "salt_minions"},
        ),
        migrations.CreateModel(
            name="Schedule",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("minion", models.CharField(max_length=128)),
                ("name", models.CharField(max_length=255)),
                ("job", models.TextField()),
            ],
        ),
        migrations.CreateModel(
            name="UserSettings",
            fields=[
                (
                    "user",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        primary_key=True,
                        related_name="user_settings",
                        serialize=False,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                ("token", models.CharField(max_length=40)),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("max_notifs", models.PositiveIntegerField(default=10)),
                ("notifs_created", models.BooleanField(default=False)),
                ("notifs_published", models.BooleanField(default=False)),
                ("notifs_returned", models.BooleanField(default=True)),
                ("notifs_event", models.BooleanField(default=False)),
                ("salt_permissions", models.TextField()),
            ],
            options={"db_table": "user_settings"},
        ),
        migrations.CreateModel(
            name="MinionsCustomFields",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255)),
                ("value", models.TextField()),
                ("function", models.CharField(max_length=255)),
                (
                    "minion",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="custom_fields",
                        to="api.Minions",
                    ),
                ),
            ],
            options={"db_table": "minions_custom_fields"},
        ),
    ]



from django.contrib.auth.models import User
from rest_framework import serializers
from rest_framework_simplejwt.serializers import TokenObtainPairSerializer
from rest_framework_simplejwt.views import TokenObtainPairView

from .models import (
    SaltReturns,
    SaltEvents,
    Keys,
    Minions,
    MinionsCustomFields,
    Conformity,
    UserSettings,
    Functions,
    Schedule,
)


class SaltReturnsSerializer(serializers.ModelSerializer):
    user = serializers.CharField()
    arguments = serializers.CharField()
    success = serializers.BooleanField(source="success_bool")

    class Meta:
        model = SaltReturns
        fields = "__all__"


class EventsSerializer(serializers.ModelSerializer):
    class Meta:
        model = SaltEvents
        fields = "__all__"


class KeysSerializer(serializers.ModelSerializer):
    class Meta:
        model = Keys
        fields = "__all__"


class MinionsCustomFieldsSerializer(serializers.ModelSerializer):
    class Meta:
        model = MinionsCustomFields
        fields = ("name", "function", "value")


class FunctionsSerializer(serializers.ModelSerializer):
    class Meta:
        model = Functions
        fields = "__all__"


class MinionsSerializer(serializers.ModelSerializer):
    last_job = serializers.DateTimeField(source="last_job.alter_time", default=None)
    last_highstate = serializers.DateTimeField(
        source="last_highstate.alter_time", default=None
    )
    conformity = serializers.BooleanField()
    custom_fields = MinionsCustomFieldsSerializer(many=True, read_only=True)

    class Meta:
        model = Minions
        fields = "__all__"


class ConformitySerializer(serializers.ModelSerializer):
    class Meta:
        model = Conformity
        fields = "__all__"


class UserSettingsSerializer(serializers.ModelSerializer):
    class Meta:
        model = UserSettings
        fields = "__all__"


class UsersSerializer(serializers.ModelSerializer):
    user_settings = UserSettingsSerializer(read_only=True)

    def create(self, validated_data):
        print(validated_data)
        for param in ["is_active", "groups", "user_permissions"]:
            del validated_data[param]
        user = User(**validated_data)
        user.set_password(validated_data["password"])
        user.save()
        return user

    class Meta:
        model = User
        fields = "__all__"
        extra_kwargs = {"password": {"write_only": True}}


class ScheduleSerializer(serializers.ModelSerializer):
    class Meta:
        model = Schedule
        fields = "__all__"


class MyTokenObtainPairSerializer(TokenObtainPairSerializer):
    def validate(self, attrs):
        data = super().validate(attrs)
        refresh = self.get_token(self.user)
        data["refresh"] = str(refresh)
        data["access"] = str(refresh.access_token)

        # Add extra responses here
        data["username"] = self.user.username
        data["id"] = self.user.id
        data["email"] = self.user.email
        return data



import binascii
import json
import os

from django.contrib.auth.models import User
from django.db.models import Q

from django.db import models


class FindJobManager(models.Manager):
    def get_queryset(self):
        return super().get_queryset().exclude(fun="saltutil.find_job")


class Jids(models.Model):
    jid = models.CharField(primary_key=True, db_index=True, max_length=255)
    load = models.TextField()

    def loaded_load(self):
        return json.loads(self.load)

    def user(self):
        if "user" in self.loaded_load():
            return self.loaded_load()["user"]
        return ""

    class Meta:
        managed = False
        db_table = "jids"
        app_label = "api"


class SaltReturns(models.Model):
    fun = models.CharField(max_length=50, db_index=True)
    jid = models.CharField(max_length=255, db_index=True)
    # Field renamed because it was a Python reserved word.
    return_field = models.TextField(db_column="return")
    id = models.CharField(max_length=255, primary_key=True)
    success = models.CharField(max_length=10)
    full_ret = models.TextField()
    alter_time = models.DateTimeField()

    objects = FindJobManager()

    def loaded_ret(self):
        return json.loads(self.full_ret)

    def user(self):
        # TODO: find a better way?
        return Jids.objects.get(jid=self.jid).user()

    def arguments(self):
        ret = self.loaded_ret()
        if "fun_args" in ret and ret["fun_args"]:
            return " ".join(str(i) for i in ret["fun_args"])
        return ""

    def success_bool(self):
        ret = self.loaded_ret()
        if "success" in ret:
            return ret["success"]
        if "return" in ret:
            if "success" in ret["return"]:
                return ret["return"]["success"]
            if "result" in ret["return"]:
                return ret["return"]["result"]
        return self.jid

    class Meta:
        managed = False
        db_table = "salt_returns"
        app_label = "api"


class SaltEvents(models.Model):
    id = models.BigAutoField(primary_key=True)
    tag = models.CharField(max_length=255, db_index=True)
    data = models.TextField()
    alter_time = models.DateTimeField()
    master_id = models.CharField(max_length=255)

    class Meta:
        managed = False
        db_table = "salt_events"
        app_label = "api"


# Alcali custom.
class Functions(models.Model):
    name = models.CharField(max_length=255)
    type = models.CharField(max_length=255)
    description = models.TextField()

    def __str__(self):
        return "{}".format(self.name)

    class Meta:
        db_table = "salt_functions"
        app_label = "api"


class Minions(models.Model):
    minion_id = models.CharField(max_length=128, null=False, blank=False)
    grain = models.TextField()
    pillar = models.TextField()

    def loaded_grain(self):
        return json.loads(self.grain)

    def loaded_pillar(self):
        return json.loads(self.pillar)

    def last_job(self):
        return (
            SaltReturns.objects.filter(id=self.minion_id)
            .order_by("-alter_time")
            .first()
        )

    def last_highstate(self):
        # Get all potential jobs.
        states = SaltReturns.objects.filter(
            Q(fun="state.apply") | Q(fun="state.highstate"), id=self.minion_id
        )
        states = sorted(states, key=lambda x: x.jid, reverse=True)

        # Remove jobs with arguments.
        for state in states:
            if (
                not state.loaded_ret()["fun_args"]
                or state.loaded_ret()["fun_args"][0] == {"test": True}
                or state.loaded_ret()["fun_args"][0] == "test=True"
            ):
                return state
        return None

    def conformity(self):
        last_highstate = self.last_highstate()
        if not last_highstate:
            return None
        highstate_ret = last_highstate.loaded_ret()
        for state in highstate_ret["return"]:
            # Flat out error(return is a string)
            if isinstance(highstate_ret["return"], list):
                return False
            # One of the state is not ok
            if not highstate_ret["return"][state]["result"]:
                return False
        return True

    def custom_conformity(self, fun, *args):

        # First, filter with fun.
        jobs = SaltReturns.objects.filter(fun=fun, id=self.minion_id).order_by(
            "-alter_time"
        )
        if not jobs:
            return False
        if args:
            for job in jobs:
                ret = job.loaded_ret()
                # if provided args are the same.
                if not list(
                    set(args) ^ {i for i in ret["fun_args"] if isinstance(i, str)}
                ):
                    return ret["return"]
        # If no args or kwargs, just return the first job.
        else:
            job = jobs.first()
            return job.loaded_ret()["return"]

    def __str__(self):
        return "{}".format(self.minion_id)

    class Meta:
        db_table = "salt_minions"
        app_label = "api"


class Keys(models.Model):
    KEY_STATUS = (
        ("accepted", "accepted"),
        ("rejected", "rejected"),
        ("denied", "denied"),
        ("unaccepted", "unaccepted"),
    )
    minion_id = models.CharField(max_length=255)
    pub = models.TextField(blank=True)
    status = models.CharField(max_length=64, choices=KEY_STATUS)

    def __str__(self):
        return "{}".format(self.minion_id)

    class Meta:
        # TODO add constraints (only one accepted per minion_id)
        db_table = "salt_keys"
        app_label = "api"


class MinionsCustomFields(models.Model):
    name = models.CharField(max_length=255)
    value = models.TextField()
    minion = models.ForeignKey(
        Minions, related_name="custom_fields", on_delete=models.CASCADE
    )
    function = models.CharField(max_length=255)

    def __str__(self):
        return "{}: {}".format(self.name, self.function)

    class Meta:
        db_table = "minions_custom_fields"
        app_label = "api"


class Schedule(models.Model):
    minion = models.CharField(max_length=128, null=False, blank=False)
    name = models.CharField(max_length=255, blank=False, null=False)
    job = models.TextField()

    def loaded_job(self):
        return json.loads(self.job)

    class Meta:
        app_label = "api"


def generate_key():
    return binascii.hexlify(os.urandom(20)).decode()


class UserSettings(models.Model):
    """
    The default authorization token model.
    """

    user = models.OneToOneField(
        User, primary_key=True, related_name="user_settings", on_delete=models.CASCADE
    )
    token = models.CharField(max_length=40)
    created = models.DateTimeField(auto_now_add=True)
    max_notifs = models.PositiveIntegerField(default=10)
    notifs_created = models.BooleanField(default=False)
    notifs_published = models.BooleanField(default=False)
    notifs_returned = models.BooleanField(default=True)
    notifs_event = models.BooleanField(default=False)
    salt_permissions = models.TextField()

    def generate_token(self):
        self.token = generate_key()
        self.save()

    class Meta:
        db_table = "user_settings"
        app_label = "api"

    def save(self, *args, **kwargs):
        if not self.token:
            self.token = generate_key()
        return super(UserSettings, self).save(*args, **kwargs)

    def __str__(self):
        return str(self.user)


class Conformity(models.Model):
    name = models.CharField(max_length=255)
    function = models.CharField(max_length=255)

    class Meta:
        db_table = "conformity"
        app_label = "api"



# from datetime import datetime
#
# import salt.config
# from salt.client import LocalClient
# from salt.exceptions import SaltReqTimeoutError
# from salt.runner import RunnerClient
# from salt.wheel import WheelClient
#
# from alcali.web.models.alcali import Minions
#
# master_conf = '/home/matt/perso/alcali/alcali/tests/conf/etc/salt/master'
# opts = salt.config.master_config(master_conf)
#
#
# def get_jobs(refresh=False):
#     if refresh:
#         try:
#             runner = RunnerClient(opts)
#             api_ret = runner.cmd(fun='jobs.list_jobs')
#             for job, value in api_ret.items():
#                 value['StartTime'] = datetime.strptime(value['StartTime'],
#                                                        '%Y, %b %d %H:%M:%S.%f')
#                 Jobs.objects.update_or_create(job_id=job,
#                                               defaults={k.lower().replace('-', '_'): v
#                                                         for k, v in value.items()})
#         except SaltReqTimeoutError:
#             print('totootot')
#     ret = Jobs.objects.all().values()
#
#     return ret
#
#
# def get_job_detail(jid):
#     runner = RunnerClient(opts)
#     ret = runner.cmd(fun='jobs.print_job', arg=[jid])
#     return ret
#
#
# def get_keys(refresh=False):
#     if refresh:
#         try:
#             wheel = WheelClient(opts)
#             ret = wheel.cmd(fun='key.list_all')
#             for minion in ret['minions']:
#                 finger = wheel.cmd(fun='key.finger', arg=[minion])
#                 finger = finger['minions'][minion]
#                 Minions.objects.update_or_create(minion_id=minion,
#                                                  defaults={'key_status': 'accepted',
#                                                         'key_finger': finger})
#         except SaltReqTimeoutError:
#             print("Connection failed!!!\n\n")
#             pass
#     ret = Minions.objects.all().values()
#
#     return ret
#
#
# def get_grains(minion_id):
#     local = LocalClient(c_path=master_conf)
#     ret = local.cmd(minion_id, 'grains.items')
#     return ret
#
#
# def run_job(tgt, fun, args, kwargs=None):
#     local = LocalClient(c_path=master_conf)
#     ret = local.cmd(tgt,
#                     fun,
#                     args,
#                     kwarg=kwargs
#                     )
#     return ret



import pytest


def test_user_settings_attr(django_user_model):
    """
    All user_settings are added on create_user.
    :return:
    """
    username = "user1"
    password = "verystrongpasswordindeed"
    user1 = django_user_model.objects.create_user(username=username, password=password)
    assert hasattr(user1, "user_settings")
    attr_list = [
        "token",
        "created",
        "notifs_created",
        "notifs_published",
        "notifs_returned",
        "notifs_event",
    ]
    for attr in attr_list:
        assert hasattr(user1.user_settings, attr)


@pytest.mark.django_db
def test_conform_highstate(minion, highstate):
    highstate = highstate()
    assert minion.conformity()
    assert highstate.success_bool()


@pytest.mark.django_db
def test_not_present_highstate(minion):
    assert minion.conformity() is None


@pytest.mark.django_db
def test_not_conform_highstate(minion, highstate_diff):
    ret = highstate_diff
    assert minion.conformity() is False


@pytest.mark.django_db
def test_failed_highstate(minion, highstate_failed):
    assert minion.conformity() is False


@pytest.mark.django_db
def test_custom_conformity_false(minion_master):
    assert minion_master.custom_conformity("cmd.run", "alcali --version") == False


@pytest.mark.django_db
def test_custom_conformity(minion_master, alcali_version_state, version_conformity):
    assert minion_master.custom_conformity("cmd.run", "alcali --version") == "2019.2.2"


@pytest.mark.django_db
def test_salt_return_job_args(jobs_arguments):
    assert jobs_arguments.arguments() == "foo 1 bar=baz"
